@@ -9,13 +9,7 @@
 
 **********************************************************************/
 
-#include "ruby/ruby.h"
-#include "ruby/io.h"
-#include "internal.h"
-#include "ruby/st.h"
-#include "ruby/util.h"
-#include "encindex.h"
-#include "id_table.h"
+#include "ruby/internal/config.h"
 
 #include <math.h>
 #ifdef HAVE_FLOAT_H
@@ -25,6 +19,24 @@
 #include <ieeefp.h>
 #endif
 
+#include "encindex.h"
+#include "id_table.h"
+#include "internal.h"
+#include "internal/array.h"
+#include "internal/bignum.h"
+#include "internal/class.h"
+#include "internal/encoding.h"
+#include "internal/error.h"
+#include "internal/hash.h"
+#include "internal/object.h"
+#include "internal/struct.h"
+#include "internal/util.h"
+#include "internal/vm.h"
+#include "ruby/io.h"
+#include "ruby/ruby.h"
+#include "ruby/st.h"
+#include "ruby/util.h"
+
 #define BITSPERSHORT (2*CHAR_BIT)
 #define SHORTMASK ((1<<BITSPERSHORT)-1)
 #define SHORTDN(x) RSHIFT((x),BITSPERSHORT)
@@ -520,10 +532,13 @@ w_extended(VALUE klass, struct dump_arg *arg, int check)
 	klass = RCLASS_SUPER(klass);
     }
     while (BUILTIN_TYPE(klass) == T_ICLASS) {
-	VALUE path = rb_class_name(RBASIC(klass)->klass);
-	w_byte(TYPE_EXTENDED, arg);
-	w_unique(path, arg);
-	klass = RCLASS_SUPER(klass);
+        if (!FL_TEST(klass, RICLASS_IS_ORIGIN) ||
+                BUILTIN_TYPE(RBASIC(klass)->klass) != T_MODULE) {
+            VALUE path = rb_class_name(RBASIC(klass)->klass);
+            w_byte(TYPE_EXTENDED, arg);
+            w_unique(path, arg);
+        }
+        klass = RCLASS_SUPER(klass);
     }
 }
 
@@ -1488,13 +1503,12 @@ r_string(struct load_arg *arg)
 static VALUE
 r_entry0(VALUE v, st_index_t num, struct load_arg *arg)
 {
-    st_data_t real_obj = (VALUE)Qundef;
-    if (arg->compat_tbl && st_lookup(arg->compat_tbl, v, &real_obj)) {
-        st_insert(arg->data, num, (st_data_t)real_obj);
-    }
-    else {
-        st_insert(arg->data, num, (st_data_t)v);
+    st_data_t real_obj = (st_data_t)v;
+    if (arg->compat_tbl) {
+        /* real_obj is kept if not found */
+        st_lookup(arg->compat_tbl, v, &real_obj);
     }
+    st_insert(arg->data, num, real_obj);
     return v;
 }
 
@@ -1678,9 +1692,6 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 
 	    v = r_object0(arg, &ivar, extmod);
 	    if (ivar) r_ivar(v, NULL, arg);
-	    if (RB_TYPE_P(v, T_STRING)) {
-	        v = r_leave(v, arg);
-	    }
 	}
 	break;
 
@@ -1728,8 +1739,7 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 	    }
 	    v = r_object0(arg, 0, extmod);
 	    if (rb_special_const_p(v) || RB_TYPE_P(v, T_OBJECT) || RB_TYPE_P(v, T_CLASS)) {
-	      format_error:
-		rb_raise(rb_eArgError, "dump format error (user class)");
+                goto format_error;
 	    }
 	    if (RB_TYPE_P(v, T_MODULE) || !RTEST(rb_class_inherited_p(c, RBASIC(v)->klass))) {
 		VALUE tmp = rb_obj_alloc(c);
@@ -1740,6 +1750,9 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 	}
 	break;
 
+      format_error:
+        rb_raise(rb_eArgError, "dump format error (user class)");
+
       case TYPE_NIL:
 	v = Qnil;
 	v = r_leave(v, arg);
@@ -1808,9 +1821,7 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 
       case TYPE_STRING:
 	v = r_entry(r_string(arg), arg);
-	if (!ivp) {
-	    v = r_leave(v, arg);
-	}
+        v = r_leave(v, arg);
 	break;
 
       case TYPE_REGEXP:
@@ -2312,9 +2323,6 @@ rb_marshal_load_with_proc(VALUE port, VALUE proc)
 void
 Init_marshal(void)
 {
-#undef rb_intern
-#define rb_intern(str) rb_intern_const(str)
-
     VALUE rb_mMarshal = rb_define_module("Marshal");
 #define set_id(sym) sym = rb_intern_const(name_##sym)
     set_id(s_dump);
